import logging
import subprocess

from .base import Plugin
from ..common import *


class NotificationsPlugin(Plugin):
    """Receive notifications from phone"""
    MARK = b'nots'
    NAME = 'NotificationsPlugin'
    MAIN_CONF = dict()
    DEVICE_CONFS = dict()
    CONFIG_SCHEMA = DictEntry('rcmd.conf.json', 'Configuration for remote commands file', False, entries=(
        IntEntry('uin', 'UIN of device for which config will be applied', True, 1, 0xFFFFFFF, None),
        FileEntry('icon_path', 'Path to save notification icon', False, '/tmp/dc-icon.png', True, False),
        TemplateEntry('cmd', 'Template of notification show command',
                      False, 0, 4096, "notify-send -i '{icon}' '{title}' '{text}'", replacements=(
                Rep('icon', 'Path to saved notification icon', False),
                Rep('title', 'Title of notification', False),
                Rep('text', 'Main content of notification', False),
            ))
    ))

    def __init__(self, app, handler, device):
        super().__init__(app, handler, device)

    def main(self):
        while True:
            request = self.rpc_read()
            self.log(request)
            if request is None:
                self.log('No more requests, stop handler')
                return
            if request.method == 'notification':
                if request.params.get('event') == 'posted':
                    cmd = self.conf('cmd')
                    icon_path = self.conf('icon_path')
                    text = request.params.get('text')
                    if text is None:
                        text = ''
                    if request.params.get('packageIcon', False):
                        icon_data = self.read()
                        if icon_path:
                            try:
                                open(icon_path, 'wb').write(icon_data)
                            except Exception as e:
                                self.log(e, logging.WARNING)
                    if cmd:
                        command = cmd.format(icon=icon_path, text=text, title=request.params.get('title', 'NULL'))
                        self.log('Execute: "{}"'.format(command))
                        subprocess.call(command, shell=True)



import fnmatch
import logging

from .base import Plugin
from ..common import *


class FileTransferPlugin(Plugin):
    """Receive file from phone"""
    MARK = b'file'
    NAME = 'FileTransferPlugin'
    MAIN_CONF = dict()
    DEVICE_CONFS = dict()
    CONFIG_SCHEMA = DictEntry('file.conf.json', 'Common configuration for file transfer plugin', False, entries=(
        IntEntry('uin', 'UIN of device for which config will be applied', True, 1, 0xFFFFFFF, None),
        DirEntry('download_directory', 'Directory to save downloaded files', False, '/tmp/dconnect', True, False),
        ListEntry('shared_dirs', 'Directories shared to client', False, 0, 1073741824, (),
                  entry=DictEntry('shared_dirs[]', 'Description of shared directory', False, entries=(
                      DirEntry('path', 'Path to shared directory', False, '/tmp/dconnect', True, False),
                      StringEntry('name', 'Name using for directory instead of path', True, 0, 60, 'Shared'),
                      StringEntry('glob', 'UNIX glob to filter visible files in directory', False, 0, 1073741824, '*'),
                      IntEntry('deep', 'Recursion deep for subdirectories', False, 1, 1024, 1)
                  )))
    ))
    PART = 65532
    shared_files_index = list()

    def check_file_filter(self, path: str, glob_str: str) -> bool:
        """Check if file allowed for sharing by filter"""
        return fnmatch.fnmatch(path, glob_str)

    def shared_directory_list(self, directory: str, filter_data, max_deep, current_deep):
        """Create information node for one shared directory"""
        res = list()
        try:
            dir_list = os.listdir(directory)
        except (PermissionError, OSError) as e:
            self.log(f'Could not list content of directory "{directory}" ({e})')
            return res
        for name in dir_list:
            path = os.path.join(directory, name)
            if os.path.isdir(path):
                if current_deep < max_deep and max_deep > 0:
                    dir_list = self.shared_directory_list(path, filter_data, max_deep, current_deep + 1)
                    res.append(dict(name=name, node_type='directory', size=len(dir_list), children=dir_list))
            elif os.path.isfile(path):
                if self.check_file_filter(path, filter_data):
                    self.shared_files_index.append(path)
                    index=len(self.shared_files_index) - 1
                    res.append(dict(name=name, node_type='file', size=os.path.getsize(path), index=index))
        return res

    def shared_files_info(self) -> list:
        """Create tree structure of shared directories"""
        self.shared_files_index.clear()
        res = list()
        names = dict()
        for shared_dir in self.conf('shared_dirs'):
            path, name = shared_dir['path'], shared_dir['name']
            if not os.path.isdir(path):
                self.log('Shared directory "{}" not found'.format(path), logging.WARN)
                continue
            if name is None:
                name = os.path.basename(path)
            if name in names:
                names[name] += 1
                name += ' ({})'.format(names[name])
            else:
                names[name] = 0
            dir_list = self.shared_directory_list(path, shared_dir['glob'], shared_dir.get('deep', 0), 1)
            res.append(dict(name=name, node_type='directory', size=len(dir_list), children=dir_list))
        return res

    def handle_upload(self, request):
        """Receive and save file from client"""
        try:
            name, size = request.params['name'], request.params['size']
        except KeyError as e:
            self.log('KeyError {}'.format(e), logging.WARN)
        else:
            path = os.path.join(self.conf('download_directory'), name)
            self.log('Receiving {} bytes to file {}'.format(size, path))
            self.rpc_send(RPCResponse(request.id, dict(code=0, message='OK')))
            f = open(path, 'wb')
            wrote = 0
            while wrote < size:
                buf = self.read()
                if buf is None:
                    self.log('File receiving aborted ({} bytes received)'.format(wrote), logging.WARN)
                    return
                if len(buf) == 0:
                    req = self.rpc_read()
                    if req.method == "cancel":
                        self.log('File receiving canceled by client ({} bytes received)'.format(wrote), logging.INFO)
                        f.close()
                        self.rpc_send(RPCResponse(request.id, dict(code=1, message='Canceled')))
                        return
                wrote += len(buf)
                f.write(buf)
            f.close()
            self.log('File received ({} bytes)'.format(wrote), logging.INFO)
            self.rpc_send(RPCResponse(request.id, dict(code=0, message='OK')))

    def handle_list_shared(self, request):
        """Create shared files info and return as JSON"""
        try:
            result = self.shared_files_info()
        except Exception as e:
            self.logger.exception('[FileTransferPlugin] {}'.format(e))
            result = INTERNAL_ERROR
        self.rpc_send(RPCResponse(request.id, result))

    def handle_download(self, request):
        """Handle try of device to download file from server"""
        try:
            index, size = request.params['index'], request.params['size']
        except KeyError as e:
            self.log('KeyError {}'.format(e), logging.WARN)
        else:
            self.log('Download request is correct')
            if 0 <= index < len(self.shared_files_index):
                path = self.shared_files_index[index]
                if not os.path.isfile(path):
                    self.rpc_send(RPCResponse(request.id, dict(code=2, message='No such file')))
                    return
                file_size = os.path.getsize(path)
                if file_size != size:
                    self.rpc_send(RPCResponse(request.id, dict(code=2, message='Size mismatch')))
                    return
                self.rpc_send(RPCResponse(request.id, dict(code=0, message='OK')))
                with open(path, 'rb') as f:
                    self.log('Start file transmission')
                    while True:
                        chunk = f.read(self.PART)
                        if len(chunk) == 0:
                            break
                        self.send(chunk)
                        self.log('Sent {} bytes...'.format(len(chunk)))
            else:
                self.rpc_send(RPCResponse(request.id, dict(code=1, message='No such index: {}'.format(index))))

    def main(self):
        while True:
            request = self.rpc_read()
            self.log(request)
            if request is None:
                self.log('[FileTransferPlugin] No more requests, stop handler')
                return
            if request.method == 'list':
                self.handle_list_shared(request)
            elif request.method == 'download':
                self.handle_download(request)
            elif request.method == 'upload':
                self.handle_upload(request)



from Crypto.Hash import SHA256
from Crypto.Cipher import AES

from .jsonrpc import *
from .jsonconf import *


def derive_key(password: str) -> bytes:
    """Create 256 bit device key from password - not really secure"""
    return SHA256.new(password.encode()).digest()


def encrypt(data: bytes, key: bytes) -> bytes:
    """Encrypt data using AES in GCM mode, nonce and digest included"""
    cipher = AES.new(key, AES.MODE_GCM)
    encrypted, digest = cipher.encrypt_and_digest(data)
    return b''.join((cipher.nonce, encrypted, digest))


def decrypt(data: bytes, key: bytes) -> Optional[bytes]:
    """Decrypt data using AES in GCM mode, also check integrity"""
    cipher = AES.new(key, AES.MODE_GCM, nonce=data[:16])
    try:
        return cipher.decrypt_and_verify(data[16:-16], data[-16:])
    except (ValueError, KeyError):
        return



from .base import PluginInitializer
from .file_transfer import FileTransferPlugin
from .remote_commands import RemoteCommandsPlugin
from .notifications import NotificationsPlugin


PLUGINS = FileTransferPlugin, RemoteCommandsPlugin, NotificationsPlugin



import time
import socket
from socketserver import TCPServer, ThreadingMixIn, BaseRequestHandler

from .common import encrypt, decrypt


class DConnectThreadingTCPServer(ThreadingMixIn, TCPServer):
    """Python TCP server with link to app class"""

    def __init__(self, app, address, handler_cls):
        self.allow_reuse_address = True
        super().__init__(address, handler_cls)
        self.app = app


class DConnectHandler(BaseRequestHandler):
    """Parse header, do authentication routines, then create plugin instance to work with connection"""
    TIMEOUT = 3

    def __init__(self, request, client_address, server):
        super().__init__(request, client_address, server)
        self.salt_send = bytes((0, ) * 32)  # get_random_bytes(32)
        self.salt_recv = None
        self.sock = self.request

    def recv(self, length):
        """Receive @length bytes from socket"""
        buf = bytearray()
        left = length
        timeout = 60
        start = time.time()
        while left > 0:
            try:
                data = self.sock.recv(left)
            except socket.timeout:
                return
            buf += data
            left = length - len(buf)
            if left > 0:
                if time.time() - start > timeout:
                    return
                time.sleep(.00001)
        return None if left else bytes(buf)

    def setup(self):
        self.sock = self.request
        self.sock.settimeout(self.TIMEOUT)

    def create_header(self, dev_self, plugin_mark, source):
        """Create connection header to send to device as auth response"""
        # Header format:
        #     ver - 8B, enc - 8B, dst - 4B, src - 4B, plg - 36B
        return b''.join((b'\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0',
                         source.uin.to_bytes(4, 'big'),
                         dev_self.uin.to_bytes(4, 'big'),
                         encrypt(plugin_mark, source.key_send)))

    def finish(self):
        """Close connection and free socket address"""
        self.sock.close()

    def handle(self):
        # Header format:
        #     ver - 8B, enc - 8B, dst - 4B, src - 4B, plg - 36B
        app = self.server.app
        log = app.log
        try:
            header = self.recv(60)  # 60 - length of header
            if header is None:
                log.warning('Header receive timeout')
                return
            dst = int.from_bytes(header[16:20], 'big')
            src = int.from_bytes(header[20:24], 'big')
            if dst != app.dev.uin:
                log.warning('Destination UIN != app UIN: {} != {}'.format(dst, app.dev.uin))
                return
            source = app.dm.get(src)
            if source is None:
                log.warning('Unknown source UIN: {}'.format(src))
                return
            if source.key_recv is None:
                log.warning('No key specified for device with UIN: {}'.format(src))
                return
            plg = decrypt(header[24:], source.key_recv)
            if plg is None:
                log.warning('Incorrect password for device with UIN: {}'.format(src))
                return
            plugin = app.plugins.get(plg)
            if plugin is None:
                log.warning('Unknown plugin mark: {}'.format(plg))
                return
            response = self.create_header(app.dev, plg, source)
            log.debug('Send header response - {} bytes'.format(len(response)))
            self.sock.sendall(response)
            log.debug('Enter plugin: "{}"'.format(plugin.NAME))
            plugin(app, self, source).main()
            log.debug('Exit plugin: "{}"'.format(plugin.NAME))
        except Exception as e:
            self.server.app.log.exception(e)
            self.finish()
            raise e



import json

class RPCObject:
    """Just group of other classes"""
    __slots__ = ()


class RPCError(BaseException, RPCObject):
    """JSON-RPC 2.0 error object"""
    __slots__ = 'code', 'message', 'data'

    def __init__(self, code, message='error', data=None):
        self.code, self.message, self.data = code, message, data

    def __repr__(self):
        return '<JSON-RPC 2.0 Error [{}]: {} - "{}">'.format(self.code, self.message, self.data)

    def to_dict(self):
        """Convert to JSON-RPC 2.0 dictionary"""
        d = dict(code=self.code, message=self.message)
        if self.data is not None:
            d['data'] = self.data
        return d

    def add_data(self, data):
        """Create copy of object with data added"""
        return type(self).__call__(self.code, self.message, data)

    @classmethod
    def from_dict(cls, d):
        """Create RPCRequest object from dictionary"""
        return cls(d['code'], d['message'], d.get('data'))


PARSE_ERROR = RPCError(-32700, 'Parse error')
INVALID_REQUEST_ERROR = RPCError(-32600, 'Invalid Request')
METHOD_NOT_FOUND_ERROR = RPCError(-32601, 'Method not found')
INVALID_PARAMS_ERROR = RPCError(-32602, 'Invalid params')
INTERNAL_ERROR = RPCError(-32603, 'Internal error')
SERVER_ERROR = RPCError(-32000, 'Server error')


class RPCRequest(RPCObject):
    """JSON-RPC 2.0 request/notification object"""
    __slots__ = 'id', 'method', 'params'

    def __init__(self, method, params, id=None):
        assert isinstance(method, str), '"method" MUST be str'
        assert isinstance(params, (tuple, list, dict)) or params is None, '"params" MUST be tuple, list, dict or None'
        assert isinstance(id, (int, str)) or id is None, '"id" MUST be int, str or None'
        self.method, self.params, self.id = method, params, id

    def to_dict(self):
        """Convert to JSON-RPC 2.0 dictionary"""
        d = dict(jsonrpc='2.0', method=self.method, params=self.params)
        if self.id is not None:
            d['id'] = self.id
        return d

    @classmethod
    def from_dict(cls, d):
        """Create RPCRequest object from dictionary"""
        try:
            return cls(d['method'], d['params'], d.get('id'))
        except (KeyError, AssertionError) as e:
            raise INVALID_REQUEST_ERROR.add_data('{}: {}'.format(type(e), str(e)))


class RPCResponse(RPCObject):
    """JSON-RPC 2.0 response object"""
    __slots__ = 'id', 'result', 'error'

    def __init__(self, id, result):
        assert isinstance(id, (int, str)) or id is None, '"id" MUST be int, str or None'
        self.id = id
        if isinstance(result, RPCError):
            self.error, self.result = result, None
        else:
            self.error, self.result = None, result

    def to_dict(self):
        """Convert to JSON-RPC 2.0 dictionary"""
        d = dict(jsonrpc='2.0', id=self.id)
        if self.error is not None:
            d['error'] = self.error.to_dict()
        if self.result is not None:
            d['result'] = self.result
        return d

    @classmethod
    def from_dict(cls, d):
        """Create RPCRequest object from dictionary"""
        try:
            result = d.get('result')
            error = d.get('error')
            if (result is not None and error is not None) or (result is None and error is None):
                raise INVALID_REQUEST_ERROR.add_data('MUST contain result XOR error')
            return cls(d['id'], result if error is None else RPCError.from_dict(error))
        except (KeyError, AssertionError) as e:
            raise INVALID_REQUEST_ERROR.add_data('{}: {}'.format(type(e), str(e)))


class RPCDispatcher:
    """Get decoded requests and return results (success or error) from corresponding methods"""

    def __init__(self, methods):
        self.methods = methods if isinstance(methods, dict) else {func.__name__: func for func in methods}

    def dispatch(self, request):
        """Check if request is correct, execute RPC method and return response"""
        func = self.methods.get(request.method)
        if func is None:
            return None if request.id is None else RPCResponse(request.id, METHOD_NOT_FOUND_ERROR)
        else:
            try:
                result = func(**request.params) if isinstance(request.params, dict) else func(*request.params)
                return None if request.id is None else RPCResponse(request.id, result)
            except TypeError as e:
                return None if request.id is None else RPCResponse(request.id, INVALID_PARAMS_ERROR.add_data(str(e)))
            except BaseException as e:
                return None if request.id is None else RPCResponse(request.id, INTERNAL_ERROR.add_data(str(e)))


class RPCSerializer:
    """Methods to serialize and deserialize JSON-RPC objects"""

    def __init__(self, ensure_ascii=True, length_bytes=None, order='big', separator=b''):
        self.ensure_ascii, self.length_bytes, self.separator, self.order = ensure_ascii, length_bytes, separator, order

    def to_bytes(self, obj):
        """Serialize JSON-RPC object to bytes"""
        try:
            return json.dumps(obj.to_dict(), separators=(',', ':'), ensure_ascii=self.ensure_ascii).encode()
        except BaseException:
            packed = json.dumps(RPCResponse(obj.id, INTERNAL_ERROR),
                                separators=(',', ':'), ensure_ascii=self.ensure_ascii).encode()
            return b''.join((len(packed).to_bytes(self.length_bytes, self.order) if self.length_bytes else b'',
                            packed, self.separator))

    def from_bytes(self, raw):
        """Extract JSON-RPC objects from byte string"""
        res = list()
        try:
            data = json.loads(raw.decode())
            if not isinstance(data, list):
                data = (data, )
            for d in data:
                try:
                    if not isinstance(d, dict):
                        res.append(INVALID_REQUEST_ERROR.add_data('Not object'))
                    if 'jsonrpc' not in d:
                        res.append(INVALID_REQUEST_ERROR.add_data('No "jsonrpc" key'))
                    if d['jsonrpc'] != '2.0':
                        res.append(INVALID_REQUEST_ERROR.add_data('JSON-RPC version != 2.0'))
                    if 'method' in d:
                        res.append(RPCRequest.from_dict(d))
                    elif 'result' in d or 'error' in d:
                        res.append(RPCResponse.from_dict(d))
                    else:
                        res.append(INVALID_REQUEST_ERROR.add_data('Not request or response'))
                except RPCError as e:
                    res.append(e)
                except BaseException as e:
                    res.append(SERVER_ERROR)
        except json.JSONDecodeError:
            res.append(PARSE_ERROR.add_data('JSON error'))
        except UnicodeDecodeError:
            res.append(PARSE_ERROR.add_data('UTF-8 error'))
        except BaseException:
            res.append(SERVER_ERROR)
        return res



import glob
import logging

from ..common import *


class PluginInitializer:
    """Facility to set some static variables for plugin classes"""

    def __init__(self, plugin_dir, log, cls):
        self.plugin_dir, self.log, self.cls = plugin_dir, log, cls

    def init_plugin(self):
        """Load plugin configurations and set up class to using in app"""
        name = self.cls.NAME
        mark = self.cls.MARK.decode('ascii')
        main_conf_path = os.path.join(self.plugin_dir, '{}.conf.json'.format(mark))
        res = ConfigLoader(main_conf_path, self.cls.CONFIG_SCHEMA, True).load()
        if isinstance(res, str):
            self.log.error(f'Plugin "{name}" init fail: {res}')
            return False
        self.cls.MAIN_CONF = res
        for path in glob.glob(os.path.join(self.plugin_dir, '*.{}.conf.json'.format(mark))):
            res = ConfigLoader(path, self.cls.CONFIG_SCHEMA, False).load()
            if isinstance(res, str):
                self.log.warning(f'Plugin "{name}" dev conf fail: {res}')
            else:
                if res.get('device') is not None:
                    self.cls.DEVICE_CONFS[res['device']] = res
                else:
                    self.log.warning(f'Plugin "{name}" dev conf fail: no UIN set in device config')
        try:
            return self.cls.post_init()
        except Exception as e:
            self.log.error('Initialization error for plugin {}'.format(name))
            self.log.exception(e)
        return False


class Plugin:
    """Base plugin class"""
    MARK = b'\0\0\0\0'
    NAME = ''
    CONF_SCHEMA = None
    DEFAULT_CONF = dict(device=None)
    MAIN_CONF = dict()
    DEVICE_CONFS = dict()

    def __init__(self, app, handler, device):
        self.app, self.logger, self.handler, self.sock, self.device = app, app.log, handler, handler.sock, device

    def log(self, message, level=logging.INFO):
        """Make log record with specified level"""
        self.logger.log(level, '[{}] {}'.format(self.NAME, message))

    @classmethod
    def post_init(cls):
        """Plugin-specific initialization, this function is called by plugin manager in the end of plugin init"""
        return True

    def conf(self, path):
        """Get value from config using path - key or key sequence.
        If uin is integer - try get from device-specific conf"""
        uin = self.device.uin
        conf = self.DEVICE_CONFS.get(uin, self.MAIN_CONF)
        if not isinstance(path, (tuple, list)):
            path = (path, )
        node = conf
        for i in path:
            try:
                node = node[i]
            except (KeyError, IndexError):
                return
            except TypeError:
                self.log('TypeError in plugin {} while config value {}'.format(self.NAME, path), logging.ERROR)
                return
        return node

    def read(self) -> Optional[bytes]:
        """Read message from socket"""
        size_raw = self.handler.recv(4)
        if size_raw is None:
            return None
        encrypted = self.handler.recv(int.from_bytes(size_raw, 'big'))
        if encrypted is None:
            return
        # key = SHA256.new(self.device.key_recv + self.handler.salt_recv)
        key = self.device.key_recv
        return decrypt(encrypted, key)

    def send(self, buf: bytes):
        """Send message to socket"""
        # key = SHA256.new(self.device.key_send + self.handler.salt_send)
        self.sock.sendall((len(buf) + 32).to_bytes(4, 'big') + encrypt(buf, self.device.key_send))

    def rpc_read(self) -> Optional[RPCRequest]:
        """Read JSON-RPC 2.0 requests/notifications"""
        request_raw = self.read()
        if request_raw is None:
            return
        self.log(request_raw, logging.DEBUG)
        try:
            return RPCRequest.from_dict(json.loads(request_raw.decode()))
        except BaseException as e:
            self.log(e, logging.WARNING)

    def rpc_send(self, obj):
        """Send JSON-RPC 2.0 response, notification or request"""
        try:
            serialized_obj = json.dumps(obj.to_dict())
            self.send(serialized_obj.encode())
            self.log("Sent: {}".format(serialized_obj), logging.DEBUG)
        except BaseException as e:
            self.log(e, logging.WARNING)

    def main(self):
        """Entry point for plugins called in TCP handler"""
        raise NotImplementedError



"""Simple UDP server to response server-search requests from devices"""

import json

from socketserver import BaseRequestHandler


class ServerSearchHandler(BaseRequestHandler):
    """Handle server search UDP requests from devices"""

    def handle(self):
        app, log, ip, socket = self.server.app, self.server.app.log, self.client_address[0], self.request[1]
        try:
            request = json.loads(self.request[0].decode())
            plugin, action = request['plugin'], request['action']
            uin, name, role = request['uin'], request['name'], request['role']
        except UnicodeDecodeError:
            log.warning('Unicode decoding error in UDP request')
        except json.JSONDecodeError as e:
            log.warning('JSON decoding error in UDP request: {}'.format(e))
        except KeyError as e:
            log.warning('Key not found in JSON (UDP request): {}'.format(e))
        else:
            if plugin == 'search' and action == 'request':
                app.dm.update_device(uin, ip, name, role)
                info = app.conf['self']
                response = json.dumps(dict(plugin='search', action='response', role='server',
                                           uin=info['uin'], name=info['name']))
                res = socket.sendto(response.encode(), self.client_address)



import os
import sys
import time
import atexit
import signal


class Daemon:
    """Set of daemon functions"""

    def __init__(self):
        self.pidfile = self.pidfile_path()

    @staticmethod
    def pidfile_path():
        """Try to determine pidfile path"""
        os.getuid()
        run_dir = os.path.join('/', 'var', 'run', 'user', str(os.getuid()))
        if os.path.isdir(run_dir):
            return os.path.join(run_dir, 'dcnnt.pid')
        else:
            return os.path.join(os.environ['HOME'], '.dcnnt.pid')

    def daemonize(self):
        """Deamonize class. UNIX double fork mechanism."""
        try:
            pid = os.fork()
            if pid > 0:
                # exit first parent
                sys.exit(0)
        except OSError as err:
            sys.stderr.write('fork #1 failed: {0}\n'.format(err))
            sys.exit(1)
        # decouple from parent environment
        os.setsid()
        os.umask(0)
        # do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                sys.exit(0)
        except OSError as err:
            sys.stderr.write('fork #2 failed: {0}\n'.format(err))
            sys.exit(1)

        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        si = open(os.devnull, 'r')
        so = open(os.devnull, 'a+')
        se = open(os.devnull, 'a+')
        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())
        # write pidfile
        atexit.register(self.delpid)
        pid = str(os.getpid())
        with open(self.pidfile, 'w+') as f:
            f.write(pid + '\n')

    def delpid(self):
        os.remove(self.pidfile)

    def check(self):
        """Check if daemon is not already running"""
        if not os.path.isfile(self.pidfile):
            return
        with open(self.pidfile, 'r') as pf:
            try:
                pid = int(pf.read().strip())
            except ValueError:
                sys.stderr.write(f'Incorrect content of pidfile "{self.pidfile}", move to "{self.pidfile}.bak"\n')
                os.rename(self.pidfile, f'{self.pidfile}.bak')
                return
        try:
            os.kill(pid, 0)
        except ProcessLookupError:
            sys.stderr.write(f'Pidfile "{self.pidfile}" exists, but no process with PID {pid} found\n')
            return
        except Exception:
            pass
        sys.stderr.write(f'Pidfile "{self.pidfile}" and process with PID {pid} exist. '
                         f'Daemon already running?\n')
        sys.exit(1)

    def stop(self):
        """Stop the daemon."""
        pid = None
        if os.path.isfile(self.pidfile):
            with open(self.pidfile, 'r') as pf:
                try:
                    pid = int(pf.read().strip())
                except ValueError:
                    sys.stderr.write(f'Incorrect content of pidfile "{self.pidfile}"\n')
                    return
        if not pid:
            sys.stderr.write(f'Pidfile "{self.pidfile} does not exist. Daemon not running?"\n')
            return  # not an error in a restart
        # Try killing the daemon process	
        try:
            while True:
                os.kill(pid, signal.SIGINT)
                time.sleep(.5)
        except OSError as e:
            error_s = str(e.args)
            if 'No such process' in error_s:
                if os.path.exists(self.pidfile):
                    os.remove(self.pidfile)
                return pid
            else:
                print(error_s)
                sys.exit(1)

    def run(self):
        """You should override this method when you subclass Daemon.
        It will be called after the process has been daemonized by 
        start() or restart()."""




"""Management of known devices and their authentication data"""

import fnmatch

from .common import derive_key
from .common.jsonconf import *


class Device:
    """Known device representation, using to manage preferences device info and key"""

    def __init__(self, uin, name, description='', role='client', password=None):
        self.uin, self.name, self.description, self.role, self.password = uin, name, description, role, password
        self.key_send = self.key_recv = self.ip = None

    def init_keys(self, app_uin, app_password):
        """Init both encrypt end decrypt keys for device"""
        self.key_recv = derive_key(''.join(map(str, (app_uin, self.uin, app_password, self.password))))
        self.key_send = derive_key(''.join(map(str, (self.uin, app_uin, self.password, app_password))))

    def dict(self):
        """Get dictionary to dump to JSON file"""
        return dict(uin=self.uin, name=self.name, description=self.description, role=self.role, password=self.password)

    def __repr__(self):
        return 'Device(UIN={}, name={}, role={}, password={})'.format(self.uin, self.name, self.role, self.password)


class DeviceManager(dict):
    """Object to control list of known devices"""
    JSON_SCHEMA = DictEntry('*.device.json', 'Device description and login data', True, entries=(
        IntEntry('uin', 'UIN of device for which config will be applied', False, 1, 0xFFFFFFF, None),
        StringEntry('name', 'Short name of device', False, 1, 40, 'Unknown'),
        StringEntry('description', 'Verbose description of device', True, 0, 200, ''),
        StringEntry('role', 'Client/server/proxy', True, 0, 40, 'client'),
        StringEntry('password', 'Access password', True, 0, 4096, ''),
    ))
    FILENAME_TEMPLATE = '{}.device.json'

    def __init__(self, app, directory):
        super().__init__()
        self.app, self.log, self.directory = app, app.log, directory

    def find_files(self, directory):
        """Find files in directory self.directory by wildcard"""
        return fnmatch.filter(os.listdir(directory), self.FILENAME_TEMPLATE.format('*'))

    def load(self):
        """Load all items using self.directory to search"""
        if os.path.isdir(self.directory):
            file_list = self.find_files(self.directory)
            for filename in file_list:
                path = os.path.join(self.directory, filename)
                self.log.info('Loading device data from file "{}"'.format(path))
                uin_item_pair = self.load_item(path)
                if uin_item_pair:
                    self[uin_item_pair[0]] = uin_item_pair[1]
                    uin_item_pair[1].init_keys(self.app.dev.uin, self.app.dev.password)
                else:
                    self.log.warning("Couldn't load device data from file '{}'".format(filename))
        else:
            self.log.warning("Devices directory '{}' not found - creating".format(self.directory))
            try:
                os.makedirs(self.directory, exist_ok=True)
            except OSError as e:
                self.log.error(f'Could not create devices directory "{self.directory}"')
                self.log.exception(e)

    def load_item(self, path):
        """Load device file data and create device object"""
        device_dict = ConfigLoader(path, self.JSON_SCHEMA, False).load()
        if isinstance(device_dict, dict):
            return device_dict['uin'], Device(**device_dict)
        self.log.warning(device_dict)

    def dump_device(self, device):
        """Save device to JSON file, using self.directory and DEVICE_FILENAME_TEMPLATE as path (WARN: plaintext pswd)"""
        filename = self.FILENAME_TEMPLATE.format(device.uin)
        path = os.path.join(self.directory, filename)
        with open(path, 'w') as f:
            json.dump(device.dict(), f, sort_keys=True, indent=2)

    def dump(self):
        """Dump all devices to files in directory self.directory"""
        for address, device in self.items():
            self.dump_device(device)

    def update_device(self, uin, ip, name=None, role=None):
        """Update current network address for known device or add new device"""
        device = self.get(uin)
        if device is None:
            self.log.info("New device found. UIN: '{}', IP: '{}'".format(uin, ip))
            if isinstance(name, str) and isinstance(role, str):
                device = self[uin] = Device(uin=uin, name=name, role=role)
                device.ip = ip
                self.dump_device(device)
            else:
                self.log.error('No role or name presented for new device')
        else:
            device.ip = ip

    def ip(self, uin):
        """Return network address for device or None if one not found"""
        device = self.get(uin)
        return None if device is None else device.ip



import subprocess

from .base import Plugin
from ..common import *


class RemoteCommandsPlugin(Plugin):
    """Receive file from phone"""
    MARK = b'rcmd'
    NAME = 'RemoteCommandsPlugin'
    MAIN_CONF = dict()
    DEVICE_CONFS = dict()
    CONFIG_SCHEMA = DictEntry('rcmd.conf.json', 'Configuration for remote commands file', False, entries=(
        IntEntry('uin', 'UIN of device for which config will be applied', True, 1, 0xFFFFFFF, None),
        ListEntry('menu', 'List of remote commands', False, 0, 1073741824, (),
                  entry=DictEntry('menu[]', 'Description of shared directory', False, entries=(
                      StringEntry('name', 'Displayed name for remote command', False, 0, 60, 'Do nothing'),
                      StringEntry('method', 'Method to execute command', True, 0, 1024, 'shell'),
                      StringEntry('cmd', 'Remote called command itself', True, 0, 1073741824, 'true'),
                  )))
    ))
    PART = 65532

    def __init__(self, app, handler, device):
        super().__init__(app, handler, device)
        self.remote_commands = dict()
        self.remote_commands_index = list()
        for command in self.conf('menu'):
            if 'cmd' in command and 'method' in command:
                identifier = str(hash(command['cmd'] + command['method']))
                self.remote_commands[identifier] = command
            else:
                identifier = None
            self.remote_commands_index.append(
                dict(index=identifier, name=command['name'], description=command.get('description')))

    def handle_exec(self, request):
        """Run command and send message with bool execution result"""
        command = self.remote_commands.get(request.params.get('index', None))
        if command is None:
            self.rpc_send(RPCResponse(request.id, dict(result=False, message='No such command')))
        else:
            method, cmd = command.get('method'), command.get('cmd')
            if method == 'shell':
                self.log('Execute shell command: "{}"'.format(cmd))
                try:
                    subprocess.check_call(cmd, shell=True)
                except Exception as e:
                    self.logger.exception(e)
                    self.rpc_send(RPCResponse(request.id, dict(result=False, message='Failed')))
                else:
                    self.rpc_send(RPCResponse(request.id, dict(result=True, message='OK')))
            else:
                self.rpc_send(RPCResponse(request.id, dict(result=False, message='No such method')))

    def main(self):
        while True:
            request = self.rpc_read()
            self.log(request)
            if request is None:
                self.log('No more requests, stop handler')
                return
            if request.method == 'list':
                self.rpc_send(RPCResponse(request.id, self.remote_commands_index))
            elif request.method == 'exec':
                self.handle_exec(request)
